from __future__ import annotations

import json
from typing import Dict, Tuple

import sqlalchemy as sa
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship

from workspaces.capability.enums import (
    CapabilityEventType,
    CapabilityRequestState,
    CapabilityStepType,
)
from workspaces.capability.helpers import CapabilitySequence, CapabilityStep
from workspaces.capability.helpers_interfaces import CapabilitySequenceIF, ParameterIF
from workspaces.capability.schema_interfaces import (
    CapabilityExecutionIF,
    CapabilityIF,
    CapabilityRequestIF,
    CapabilityVersionIF,
)
from workspaces.products.schema_interfaces import FutureProductIF


class CapabilityEvent:
    """
    An event sent from the underlying workflow execution system.
    """

    def __init__(self, event_type: CapabilityEventType, request_id: int, execution_id: int):
        self.event_type = event_type
        self.execution_id = execution_id

    def json(self) -> str:
        return json.dumps(self.__dict__)

    def __str__(self):
        return f"<CapabilityEvent with data{self.json()}>"

    def __repr__(self):
        return f"<CapabilityEvent {self.__dict__}>"

    def __eq__(self, other):
        return self.json() == other.json()


Base = declarative_base()


class Capability(Base, CapabilityIF):
    """
    A capability, which is a particular workflow setup, intended to accept
    a certain kind of product and some parameters and produce another product
    """

    __tablename__ = "capabilities"
    name = sa.Column("capability_name", sa.String, primary_key=True)
    steps = sa.Column("capability_steps", sa.String)
    max_jobs = sa.Column("max_jobs", sa.Integer)
    enabled = sa.Column("enabled", sa.Boolean, default=True, server_default="true")
    requests = relationship("CapabilityRequest", back_populates="capability")

    @classmethod
    def from_file(cls, filename: str) -> Capability:
        """
        Method that creates an instance of a Capability from a given file

        :param filename:
        :return: Capability instance
        """
        self = cls()
        self.name, self.max_jobs, self.steps = cls.parse_capability_file(filename)
        return self

    @classmethod
    def from_json(cls, json_data: str) -> Capability:
        """
        Method that creates an instance of a Capability from given JSON

        :param json_data:
        :return: Capability instance
        """
        self = cls()
        json_dict = json.loads(json_data)
        self.name = json_dict["name"]
        self.max_jobs = json_dict["max_jobs"]
        self.steps = json_dict["steps"]
        self.enabled = json_dict.get("enabled", True)

        return self

    @staticmethod
    def parse_capability_file(
        filename: str,
    ) -> Tuple[str, int, CapabilitySequenceIF]:
        """
        Parse a file containing metadata for a capability

        :param filename:
        :return: Tuple containing capability's ID, name, max concurrent jobs, and sequence of steps
        """
        with open(filename, "r") as f:
            name, max_jobs = f.readline().split(" ")
            steps = []

            for line in f.readlines():
                step_type, step_value = line.strip().split(" ")
                steps.append(CapabilityStep(CapabilityStepType.from_string(step_type), step_value))

            return name, int(max_jobs), CapabilitySequence(steps)

    def create_request(self, parameters: ParameterIF, future_products: FutureProductIF):
        """
        Create a new request for this capability with specific options

        FIXME: Future products needs a string representation
        """
        request = CapabilityRequest(capability=self, parameters=str(parameters))
        self.requests.append(request)
        return request

    def __str__(self) -> str:
        return (
            f"Capability: {self.name}, max concurrent jobs of {self.max_jobs}"
            f"\nSequence: {self.steps}"
        )

    def __json__(self) -> Dict[str, str]:
        return {
            "name": self.name,
            "max_jobs": self.max_jobs,
            "steps": str(self.steps),
            "enabled": self.enabled,
        }


class CapabilityRequest(Base, CapabilityRequestIF):
    """
    A capability request, which couples a capability to a product, representing
    the expectation of a new product given a set of parameters
    """

    __tablename__ = "capability_requests"
    id = sa.Column("capability_request_id", sa.Integer, primary_key=True)
    state = sa.Column("state", sa.String)
    capability_name = sa.Column(
        "capability_name", sa.String, sa.ForeignKey("capabilities.capability_name")
    )
    parameters = sa.Column("parameters", sa.String)
    # FIXME: This needs to be changed to properly keep track of product locators.
    # future_products = sa.Column('future_products', sa.String)
    versions = relationship("CapabilityVersion", back_populates="request")
    capability = relationship(Capability, back_populates="requests")

    def update_state(self, state: CapabilityRequestState):
        self.state = state.name

    def __str__(self):
        return f"CapabilityRequest object: {self.__dict__}"

    def __json__(self) -> Dict[str, str]:
        return {
            "id": self.id,
            "capability_name": self.capability_name,
            "state": self.state,
            "parameters": self.parameters,
        }


class CapabilityVersion(Base, CapabilityVersionIF):
    """
    Schema representation of a capability version
    """

    __tablename__ = "capability_versions"
    capability_request_id = sa.Column(
        "capability_request_id",
        sa.Integer,
        sa.ForeignKey("capability_requests.capability_request_id"),
        primary_key=True,
    )
    version_number = sa.Column("version_number", sa.Integer, primary_key=True)
    parameters = sa.Column("parameters", sa.String)
    request = relationship(CapabilityRequest, back_populates="versions")
    executions = relationship("CapabilityExecution", back_populates="version")

    @property
    def capability(self):
        return self.request.capability


class CapabilityExecution(Base, CapabilityExecutionIF):
    """
    Schema representation of a capability request's execution record
    """

    __tablename__ = "capability_executions"

    id = sa.Column("execution_id", sa.Integer, primary_key=True)
    state = sa.Column("state", sa.String)
    capability_request_id = sa.Column(
        "capability_request_id",
        sa.Integer,
        sa.ForeignKey("capability_requests.capability_request_id"),
    )
    version_number = sa.Column("capability_version_number", sa.Integer)
    current_step = sa.Column("current_step", sa.Integer)
    steps = sa.Column("steps", sa.String)
    version = relationship(CapabilityVersion, back_populates="executions")
    current_workflow_request_id = sa.Column("current_workflow_request_id", sa.Integer)

    __table_args__ = (
        sa.ForeignKeyConstraint(
            (capability_request_id, version_number),
            [CapabilityVersion.capability_request_id, CapabilityVersion.version_number],
        ),
    )

    def __json__(self) -> dict:
        return dict(
            id=self.id,
            state=self.state,
            capability_request_id=self.capability_request_id,
            version_number=self.version_number,
            current_step=self.current_step,
            current_workflow_request_id=self.current_workflow_request_id,
        )

    def update_execution(self, workflow_request_id: int):
        self.current_workflow_request_id = workflow_request_id

    @property
    def capability(self):
        return self.version.capability

    @property
    def capability_request(self):
        return self.version.request
